"""Implementation of the various scripts available from bin.

This is structured this way because we want all the significant codes under a
single directory, for analysis, grepping and unit testing.
"""
__copyright__ = "Copyright (C) 2014, 2016  Martin Blais"
__license__ = "GNU GPLv2"
